%          +---------------------------------------------------+
%          | Department of Signal Processing and Acoustics     |
%          | Faculty of Elec., Comm. & Automation              |
%          | wooster.tkk.fi - Espoo - Finland                  |
%          +---------------------------------------------------+
%          | Mohammed Mobien                                   |
%          +---------------------------------------------------+
%          | Algorithm: Weight Extraction for FQRD-RLS alg.    |
%          +---------------------------------------------------+
%          | Ref.:      Weight Extraction of Fast QRD-RLS Alg. |
%          |            QRD-RLS Adaptive Filtering             |
%          |            Apolinrio, Jos Antonio Jr (Ed.)      |
%          |            2009, Approx. 385 p.,                  |
%          |            ISBN: 978-0-387-09733-6                |
%          +---------------------------------------------------+

clear; clc; close all; pack;
% Ininitalization of constants
K         = 1000;      % number of iterations
ENS       = 10;        % ensemble
varx      = 1;         % input signal variance
varn      = 1e-8;      % noise variance
nrcoef    = 4;         % number of coeficients (N+1)
N         = nrcoef-1;  % filter order
num       = [1];       % num and den used to generate colored input signal
den       = [1 1.2 0.81];   
lambda    = 0.99;      % forgetting factor
lambda05  = sqrt(lambda);
lambdam05 = 1/lambda05;
Wopt      = [1 0.9*sqrt(-1) 0.1 0.2*sqrt(-1)]';              
MSEIQR    = zeros(K,1);
MSEFQR    = zeros(K,1);
MSED      = zeros(K,1);
delta = 1;          
for ens=1:ENS
  disp(ens);
  % Generating input and reference (desired) signals corrupted with the measurement noise:
  x     = randn(K,1)+sqrt(-1)*randn(K,1);                             
  x     = filter(num,den,x); x=x/std(x);  x=x*sqrt(varx);
  noise = randn(K,1); noise=noise/std(noise); noise=noise*sqrt(varn);  
  d     = filter(conj(Wopt),1,x)+noise;
  %------------------------------------------------
  % Initialization of FQRD-RLS algorithm variables:
  %------------------------------------------------
  moduloef=delta;
  dfq2=zeros(N+1,1);
  dq2=zeros(N+1,1);
  ctheta=ones(N+1,1);
  stheta=zeros(N+1,1);
  cthetaprimef=ones(N+1,1);
  sthetaprimef=zeros(N+1,1);
  f=zeros(N+1,1);
  gamma=1;
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  % Extracting the Cholesky factor matrix for exact initialization
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  % Being
  xi=zeros(N+1,1); xi(1)=1;
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  % Running the weight extraction alg.
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  for k = 1:N+1
      %---------------------
      % 1. Obtaining u(k):
      %---------------------
      f(N+1) = f(N+1)/ctheta(1);
      aux = -conj(stheta(1))*f(N+1);
      for i = 1:N
          f(N+1-i) = (f(N+1-i) - (stheta(i+1))*aux)/ctheta(i+1);
          aux = ctheta(i+1)*aux - conj(stheta(i+1))*f(N+1-i);
      end
      f = lambda05*f;
      %---------------------
      % 2. Obtaining u(k+1):
      %---------------------
      aux=zeros(N+2,1);
      aux(N+2)= conj(xi(k) - f'*dfq2)/moduloef;
      for i=1:(N+1)
          aux(i)=cthetaprimef(i)*f(i)-conj(sthetaprimef(i))*aux(N+2);
          aux(N+2)=sthetaprimef(i)*f(i)+cthetaprimef(i)*aux(N+2);
      end % for i
      f=aux(2:N+2);
      Uiqr(:, k) = f;
  end
  f=zeros(N+1,1);
  % End
  %%%%%%%%%%%%%%%
  %------------------------------------------------
  % Initialization of IQRD-RLS algorithm variables:
  %------------------------------------------------
  vectorx = zeros(N+1,1);
  W       = zeros(N+1,1);  
  dq2aux  = zeros(N+1,1);    
  % CHECK THE EFFECT OF THE INITIALIZATION:  
  %delta=0.001;
  % with 1, 0.1 or 0.01 instead of 0.001 or with a random lower triang. U(k-1):   
  %U       = inv((1/delta)*randn(N+1,N+1).*hankel((e/delta)*randn(N+1,1)));
  %UmH     = inv(U)'; 
  %UmH  = zeros(N+1); for n=1:(N+1); UmH(n,N+2-n)=1/delta; end; % Reversal matrix
  UmH  = Uiqr;

  %--------------------
  % Running the filter:
  %--------------------
  for k=1:K
    if rem(k,10000)==0
      disp(k);
    end   
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % FQRD-RLS algorithm starts
    %-------------------------------------
    % 1. Obtaining dfq2(k+1) and efq1(k+1:
    %-------------------------------------
    efq1=conj(x(k));
    for i=1:N+1
      oldfriend=efq1;
      efq1=ctheta(i-1+1)*efq1-conj(stheta(i-1+1))*lambda05*dfq2(N+2-i);
      dfq2(N+2-i)=stheta(i-1+1)*oldfriend+ctheta(i-1+1)*lambda05*dfq2(N+2-i);
    end; % for i
    %--------------------------
    % 2. Obtaining ||ef(k+1)||:
    %--------------------------
    moduloef=sqrt(abs(efq1)^2+lambda*moduloef^2);
    %----------------------------
    % 3. Obtaining Qtheta'f(k+1):
    %----------------------------
    moduloef0=moduloef;
    for i=1:N+1,
        oldfriend=moduloef0;
        moduloef0=sqrt(moduloef0^2+abs(dfq2(i))^2);
        cthetaprimef(i)=oldfriend/moduloef0;
        sthetaprimef(i)=conj(dfq2(i))/moduloef0;      
    end; % for i
    %---------------------
    % 4. Obtaining f(k+1):
    %---------------------
    aux=zeros(N+2,1);
    aux(N+2)=gamma*conj(efq1)/moduloef;
    for i=1:(N+1)
        aux(i)=cthetaprimef(i)*f(i)-conj(sthetaprimef(i))*aux(N+2);
        aux(N+2)=sthetaprimef(i)*f(i)+cthetaprimef(i)*aux(N+2);
    end % for i
    f=aux(2:N+2);
    %-------------------------
    % 5. Obtaining Qteta(K+1): 
    %-------------------------
    gamma=1;
    for i=1:N+1
        stheta(i-1+1)=f(N+2-i)/gamma;
        ctheta(i-1+1)=sqrt(1-abs(stheta(i-1+1))^2);
        gamma=ctheta(i-1+1)*gamma;
    end; % for i
    %-----------------------------
    % 6. Joint Process Estimation:
    %-----------------------------
    eq1=conj(d(k));
    for i=1:N+1
        oldfriend=eq1;
        eq1=ctheta(i-1+1)*eq1-conj(stheta(i-1+1))*lambda05*dq2(N+2-i);
        dq2(N+2-i)=stheta(i-1+1)*oldfriend+ctheta(i-1+1)*lambda05*dq2(N+2-i);
    end; % for i
    %------------------------------
    % 7. Updating the output error:
    %------------------------------
    ekaposteriori=conj(eq1)*gamma;
    ekapriori=conj(eq1)/gamma;
    % FQRD-RLS algorithm ends
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % IQRD-RLS algorithm starts
    vectorx = [x(k); vectorx(1:N)];
    %--------------------------------------
    % Obtaining a(k)=lambdam05*UmH*vectorx:
    %--------------------------------------    
    ak=zeros(N+1,1);
    xaux=lambdam05*vectorx;
    for n=1:(N+1)
       for m=1:(N+2-n)
            ak(n)=ak(n)+UmH(n,m)*xaux(m);
       end
    end
    %----------------------------------
    % Obtaining Qtheta(k) and gamma(k):
    %----------------------------------
    igamma = 1; 
    for n=1:N+1   
       aux1      = sqrt(abs(igamma)^2+abs(ak(N+2-n))^2); 
       cthetaI(n) = abs(igamma)/aux1; % 
       sthetaI(n) = -(-ak(N+2-n)/igamma)*ctheta(n); % steta(n)=ak(N+2-n)/aux1;
       igamma    = aux1; % ctheta(n)*igamma+conj(stheta(n))*ak(N+2-n); 
    end
    gamma = 1/igamma;
    %---------------------------------------
    % Obtaining u(k) and updating U^{H}(k):
    %---------------------------------------
    uHaux = zeros(N+1,1);
    UmHaux=lambdam05*UmH;
    for n=1:(N+1)
        for m=1:n
            aux2            = uHaux(m);
            uHaux(m)        = cthetaI(n)*aux2-conj(sthetaI(n))*UmHaux(N+2-n,m);
            UmHaux(N+2-n,m) = sthetaI(n)*aux2+cthetaI(n)*UmHaux(N+2-n,m);
        end
    end
    u   = conj(uHaux);
    UmH = UmHaux;
    %---------------
    %Obtaining e(k):
    %---------------
    epriori = d(k)-W'*vectorx;
    %---------------------------------
    % Updating the coefficient vector:
    %---------------------------------
    W       = W-gamma*u*conj(epriori);
    % IQRD-RLS algorithm ends
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Learning curves for both algorithms
    MSEIQR(k)  =MSEIQR(k)+abs(epriori)^2;
    MSEFQR(k)  =MSEFQR(k)+abs(ekapriori)^2;
    MSED(k)    =MSED(k)+abs(ekapriori-epriori)^2;
    % MSE(k)=MSE(k)+abs(ekaposteriori)^2;
  end % for k
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  % Initialization for weight extraction
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  f=zeros(N+1,1);
  x=zeros(N+1,1); x(1)=1;
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  % Running the weight extraction alg.
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  for k = 1:N+1
    %---------------------
    % 1. Obtaining u(k):
    %---------------------
    f(N+1) = f(N+1)/ctheta(1);
    aux = -conj(stheta(1))*f(N+1);
    for i = 1:N
        f(N+1-i) = (f(N+1-i) - (stheta(i+1))*aux)/ctheta(i+1);
        aux = ctheta(i+1)*aux - conj(stheta(i+1))*f(N+1-i);
    end
    f = lambda05*f;
    %---------------------
    % 2. Obtaining u(k+1):
    %---------------------
    aux=zeros(N+2,1);
    aux(N+2)= conj(x(k) - f'*dfq2)/moduloef;
    for i=1:(N+1)
        aux(i)=cthetaprimef(i)*f(i)-conj(sthetaprimef(i))*aux(N+2);
        aux(N+2)=sthetaprimef(i)*f(i)+cthetaprimef(i)*aux(N+2);
    end % for i
    f=aux(2:N+2);
    wFQR(k) = f'*dq2;
  end
  
end % for ens
MSEFQR=MSEFQR/ENS;
MSEIQR=MSEIQR/ENS;
MSED=MSED/ENS;
%-------------------
% Simulation result:
%-------------------
Efqr=10*log10(MSEFQR);
Eiqr=10*log10(MSEIQR);
Wdiff=10*log10(abs(wFQR(:)-W).^2);
Ed=10*log10(MSED);
k=1:K;
figure(1);
plot(k,Efqr,'k'); hold on; plot(k,Eiqr,'m');
xlabel('k'); ylabel('MSE_{dB}');
title('FQRPOSB Version 2 and inverse QRD-RLS: learning curves');
figure(2);
plot(1:N+1,Wdiff,'k'); 
xlabel('coeff'); ylabel('weight diff');
title('Weight difference of IQR and FQR algorithms');
figure(3);
plot(k,Ed,'k'); 
xlabel('k'); ylabel('MSE_{dB}');
title('FQRPOSB Version 2 and inverse QRD-RLS: difference of learning curves');

%--------------------> END OF WEIGHTEXTRACTION.M <--------------------%

