% +--------------------------------+
% | Algorithm: The Inverse QRD-RLS |
% | Jos Antonio Apolinrio Jr.    |
% +--------------------------------+
clear; clc; close all; pack;
K         = 1000;      % number of iterations
ENS       = 5;         % ensemble
varx      = 1;         % input signal variance
varn      = 1e-8;      % noise variance
nrcoef    = 4;         % number of coeficients (N+1)
N         = nrcoef-1;  % filter order
num       = [1];       % num and den used to generate colored input signal
den       = [1 1.2 0.81];   
lambda    = 0.99;      % forgetting factor
lambda05  = sqrt(lambda);
lambdam05 = 1/lambda05;
Wopt      = [1 0.9*sqrt(-1) 0.1 0.2*sqrt(-1)]';              
MSE       = zeros(K,1);
for ens=1:ENS,
  disp(ens);
  % Generating input and reference (desired) signals + QRD-RLS init.:
  x     = randn(K,1)+sqrt(-1)*randn(K,1);                             
  x     = filter(num,den,x); x=x/std(x);  x=x*sqrt(varx);
  noise = randn(K,1); noise=noise/std(noise); noise=noise*sqrt(varn);  
  d     = filter(conj(Wopt),1,x)+noise;
  %----------------
  % Initialization:
  %----------------  
  vectorx = zeros(N+1,1);
  W       = zeros(N+1,1);  
  dq2aux  = zeros(N+1,1);    
  % CHECK THE EFFECT OF THE INITIALIZATION:  
  delta=0.001;
  % with 1, 0.1 or 0.01 instead of 0.001 or with a random lower triang. U(k-1):   
  %U       = inv((1/delta)*randn(N+1,N+1).*hankel((e/delta)*randn(N+1,1)));
  %UmH     = inv(U)'; 
  UmH  = zeros(N+1); for n=1:(N+1); UmH(n,N+2-n)=1/delta; end; % Reversal matrix   
  %---------
  % Running:
  %---------
  for k=1:K 
    vectorx = [x(k)
               vectorx(1:N)];
    %--------------------------------------
    % Obtaining a(k)=lambdam05*UmH*vectorx:
    %--------------------------------------    
    ak=zeros(N+1,1);
    xaux=lambdam05*vectorx;
    for n=1:(N+1)
       for m=1:(N+2-n)
            ak(n)=ak(n)+UmH(n,m)*xaux(m);
       end
    end
    %----------------------------------
    % Obtaining Qtheta(k) and gamma(k):
    %----------------------------------
    igamma = 1; 
    for n=1:N+1   
       aux1      = sqrt(abs(igamma)^2+abs(ak(N+2-n))^2); 
       ctheta(n) = abs(igamma)/aux1; % 
       stheta(n) = -(-ak(N+2-n)/igamma)*ctheta(n); % steta(n)=ak(N+2-n)/aux1;
       igamma    = aux1; % ctheta(n)*igamma+conj(stheta(n))*ak(N+2-n); 
    end
    gamma = 1/igamma;
    %---------------------------------------
    % Obtaining u(k) and updating U^{?H}(k):
    %---------------------------------------
    uHaux = zeros(N+1,1);
    UmHaux=lambdam05*UmH;
    for n=1:(N+1)
        for m=1:n
            aux2            = uHaux(m);
            uHaux(m)        = ctheta(n)*aux2-conj(stheta(n))*UmHaux(N+2-n,m);
            UmHaux(N+2-n,m) = stheta(n)*aux2+ctheta(n)*UmHaux(N+2-n,m);
        end
    end
    u   = conj(uHaux);
    UmH = UmHaux;
    %---------------
    %Obtaining e(k):
    %---------------
    epriori = d(k)-W'*vectorx;
    %---------------------------------
    % Updating the coefficient vector:
    %---------------------------------
    W       = W-gamma*u*conj(epriori);
    MSE(k)  = MSE(k)+abs(epriori)^2;
  end; % for k
end; % for ens
MSE      = MSE/ENS;
MSEdB    = 10*log10(MSE);
MSEmindB = 10*log10(varn)*ones(K,1);
%--------------------
% Simulation results:
%--------------------
xscale=1:K;
plot(xscale,MSEdB); hold;
xlabel('k'); ylabel('E[|e(k)|^2]_{dB}');
title('Learning curve (a priori error) of the Inverse QRD-RLS');
plot(xscale,MSEmindB,'--r')
%---------------------------------------
