% +--------------------------------+
% | Algorithm: FQR_POS_F           |
% | Jos Antonio Apolinrio Jr.    |
% +--------------------------------+
clear; clc; close all; pack;
K         = 1000;      % number of iterations
ENS       = 10;        % ensemble
varx      = 1;         % input signal variance
varn      = 1e-8;      % noise variance
nrcoef    = 4;         % number of coeficients (N+1)
N         = nrcoef-1;  % filter order
num       = [1];       % num and den used to generate colored input signal
den       = [1 1.2 0.81];   
lambda    = 0.99;      % forgetting factor
lambda05  = sqrt(lambda);
lambdam05 = 1/lambda05;
Wopt      = [1 0.9 0.1 0.2]';
MSE       = zeros(K,1);
delta = 1e-4;
for ens=1:ENS,
  disp(ens)
  % Generating input and reference (desired) signals:
  x     = randn(K,1);
  x     = filter(num,den,x); x=x/std(x);  x=x*sqrt(varx);
  noise = randn(K,1); noise=noise/std(noise); noise=noise*sqrt(varn);  
  d     = filter(Wopt,1,x)+noise;
  %----------------
  % Initialization:
  %----------------   
  dfq2=zeros(N+1,1);
  dq2=zeros(N+1,1);
  moduloef=delta;
  f=zeros(N+1,1);
  ctheta=ones(N+1,1);
  cthetaprimeb=ones(N+1,1);
  stheta=zeros(N+1,1);
  sthetaprimeb=zeros(N+1,1);
  gamma=1;
  %-------------------
  % Runing the filter:
  %-------------------
  for k=1:K
    if rem(k,10000)==0
      disp(k);
    end
    %------------------------
    % 1. Obtaining efq1(k+1):
    %------------------------
    efq1=x(k);
    for i=1:N+1
      oldfriend=dfq2(i);
      dfq2(i)=stheta(i)*efq1+ctheta(i)*lambda05*dfq2(i);
      efq1=ctheta(i)*efq1-stheta(i)*lambda05*oldfriend;
    end % do for i
    %--------------------------
    % 2. Obtaining Qthetaf(k+1)
    %--------------------------
    moduloefold=moduloef;
    % updating ||ef(k)||
    moduloef=sqrt(lambda*moduloef^2+(efq1)^2);
    cthetaf=lambda05*moduloefold/moduloef;
    sthetaf=efq1/moduloef; 
    %---------------------
    % 3. Obtaining c(k+1):
    %---------------------    
    c=[1
       zeros(N+1,1)];
    % c=Qthetab'*c;
    for i=1:(N+1)
      c(N+3-i)=-sthetaprimeb(N+2-i)*c(1);
      c(1)=cthetaprimeb(N+2-i)*c(1);
    end
    caux=[0
          c];
    % caux(1:N+2)=Qtheta(k)*caux(1:N+2)
    for i=1:(N+1)
      oldfriend=caux(i+1);
      caux(i+1)=stheta(i)*caux(1)+ctheta(i)*caux(i+1);
      caux(1)=ctheta(i)*caux(1)-stheta(i)*oldfriend;
    end;
    % caux=Qthetaf(k)*caux
    oldfriend=caux(1);
    caux(1)=cthetaf*caux(1)-sthetaf*caux(N+3);
    caux(N+3)=sthetaf*oldfriend+cthetaf*caux(N+3);
    c=caux(2:(N+3));
    %--------------------------------
    % 4. Obtaining Qthetaprimeb(k+1):
    %--------------------------------
    b=c(1);
    for i=1:(N+1)
      oldfriend=b;
      b=sqrt(b^2+(c(i+1))^2);
      cthetaprimeb(i)=oldfriend/b; 
      sthetaprimeb(i)=-c(i+1)/b;   
    end    
    %---------------------
    % 5. Obtaining f(k+1):
    %---------------------
    fNplus2=[f
             gamma*sthetaf]; % or gamma*efq1/moduloef; since sthetaf=efq1/moduloef; 
    aux=fNplus2(1);
    for i=1:N+1
      f(i)=sthetaprimeb(i)*aux+cthetaprimeb(i)*fNplus2(i+1);
      aux=cthetaprimeb(i)*aux-sthetaprimeb(i)*fNplus2(i+1);
    end % for i
    %------------------------
    % 6. Obtendo Qtheta(k+1):
    %------------------------
    gamma=1;
    for i=1:N+1
       stheta(i)=f(i)/gamma;
       ctheta(i)=sqrt(1-stheta(i)^2);
       gamma=ctheta(i)*gamma;
    end; % for i
    %-----------------------------
    % 7. Joint Process Estimation:
    %-----------------------------
    eq1=d(k);
    for i=1:N+1
        oldfriend=eq1;
        eq1=ctheta(i)*eq1-stheta(i)*lambda05*dq2(i);
        dq2(i)=stheta(i)*oldfriend+ctheta(i)*lambda05*dq2(i);        
    end; % for i    
    %------------------------------
    % 8. Updating the output error:
    %------------------------------
    ekaposteriori=eq1*gamma;
    ekapriori=eq1/gamma;
    MSE(k)=MSE(k)+abs(ekapriori)^2;
    % MSE(k)=MSE(k)+abs(ekaposteriori)^2;    
  end; % for k
end; % for ens
MSE=MSE/ENS;
%-------------------
% Simulation result:
%-------------------
E=10*log10(MSE);
k=1:K;
plot(k,E,'k');
xlabel('k'); ylabel('MSE_{dB}');
title('FQR\_POS\_F: learning curve');
%--------------------> END OF FQRPOSF.M <--------------------%

