% +--------------------------------+
% | Algorithm: FQR_PRI_B version 2 |
% | Jos Antonio Apolinrio Jr.    |
% +--------------------------------+
clear; clc; close all; pack;
K         = 1000;      % number of iterations
ENS       = 10;        % ensemble
varx      = 1;         % input signal variance
varn      = 1e-8;      % noise variance
nrcoef    = 4;         % number of coeficients (N+1)
N         = nrcoef-1;  % filter order
num       = [1];       % num and den used to generate colored input signal
den       = [1 1.2 0.81];   
lambda    = 0.99;      % forgetting factor
lambda05  = sqrt(lambda);
lambdam05 = 1/lambda05;
Wopt      = [1 0.9*sqrt(-1) 0.1 0.2*sqrt(-1)]';              
MSE       = zeros(K,1);
delta = 1e-6;
for ens=1:ENS
  disp(ens); 
  % Generating input and reference (desired) signals:
  x     = randn(K,1)+sqrt(-1)*randn(K,1);                             
  x     = filter(num,den,x); x=x/std(x);  x=x*sqrt(varx);
  noise = randn(K,1); noise=noise/std(noise); noise=noise*sqrt(varn);  
  d     = filter(conj(Wopt),1,x)+noise;
  %----------------
  % Initialization:
  %----------------
  moduloef=delta;
  dfq2=zeros(N+1,1);
  dq2=zeros(N+1,1);
  ctheta=ones(N+1,1);
  stheta=zeros(N+1,1);
  cthetaprimef=ones(N+1,1);
  sthetaprimef=zeros(N+1,1);
  a=zeros(N+2,1);
  gamma=1;
  %--------------------
  % Running the filter:
  %--------------------
  for k=1:K
    if rem(k,10000)==0
      disp(k);
    end
    %-------------------------------------
    % 1. Obtaining dfq2(k+1) and efq1(k+1:
    %-------------------------------------
    efq1=conj(x(k));
    for i=1:N+1
      oldfriend=efq1;
      efq1=ctheta(i-1+1)*oldfriend-conj(stheta(i-1+1))*lambda05*dfq2(N+2-i);
      dfq2(N+2-i)=stheta(i-1+1)*oldfriend+ctheta(i-1+1)*lambda05*dfq2(N+2-i);
    end; % for i
    %---------------------
    % 2. Obtaining a(k+1):
    %---------------------
    aux=efq1/(gamma*lambda05*moduloef);
    for i=1:(N+1)
        a(i-1+1)=cthetaprimef(N+1-i+1)*a(i+1)-(sthetaprimef(N+1-i+1))*aux;
        aux=conj(sthetaprimef(N+1-i+1))*a(i+1)+cthetaprimef(N+1-i+1)*aux;
    end % for i
    a(N+1+1)=aux;
    %--------------------------
    % 3. Obtaining ||ef(k+1)||:
    %--------------------------
    moduloef=sqrt(abs(efq1)^2+lambda*moduloef^2);
    %----------------------------
    % 4. Obtaining Qtheta'f(k+1): 
    %----------------------------
    moduloef0=moduloef;
    for i=1:N+1,
      oldfriend=moduloef0;
      moduloef0=sqrt(moduloef0^2+abs(dfq2(i))^2);
      cthetaprimef(N+1-i+1)=oldfriend/moduloef0;
      sthetaprimef(N+1-i+1)=conj(dfq2(i))/moduloef0;
    end; % for i
    %--------------------------
    % 5. Obtaining Qtheta(K+1):
    %--------------------------
    invgamma=1;
    for i=1:N+1
      oldfriend=invgamma;
      invgamma=sqrt(invgamma^2+abs(a(N+2-i+1))^2);
      ctheta(i-1+1)=oldfriend/invgamma;
      stheta(i-1+1)=conj(a(N+2-i+1))/invgamma;
    end; % for i
    gamma=1/invgamma;
    %-----------------------------
    % 6. Joint Process Estimation:
    %-----------------------------
    eq1=conj(d(k));
    for i=1:N+1
      oldfriend=eq1;
      eq1=ctheta(i-1+1)*eq1-conj(stheta(i-1+1))*lambda05*dq2(N+2-i);
      dq2(N+2-i)=stheta(i-1+1)*oldfriend+ctheta(i-1+1)*lambda05*dq2(N+2-i);
    end; % for i
    %------------------------------
    % 7. Updating the output error:
    %------------------------------    
    ekaposteriori=conj(eq1)*gamma;
    ekapriori=conj(eq1)/gamma;
    MSE(k)=MSE(k)+abs(ekapriori)^2;
    % MSE(k)=MSE(k)+abs(ekaposteriori)^2;
  end % for k
end % for ens
MSE=MSE/ENS;
%-------------------
% Simulation result:
%-------------------
E=10*log10(MSE);
k=1:K;
plot(k,E,'k');
xlabel('k'); ylabel('MSE_{dB}');
title('FQRPRIB Version 2: learning curve');
%--------------------> END OF FQRPRIB2.M <--------------------%
