% +--------------------------------+
% | Algorithm: FQR_PRI_F           |
% | Jos Antonio Apolinrio Jr.    |
% +--------------------------------+
clear; clc; close all; pack;
K         = 1000;      % number of iterations
ENS       = 10;        % ensemble
varx      = 1;         % input signal variance
varn      = 1e-8;      % noise variance
nrcoef    = 4;         % number of coeficients (N+1)
N         = nrcoef-1;  % filter order
num       = [1];       % num and den used to generate colored input signal
den       = [1 1.2 0.81];   
lambda    = 0.99;      % forgetting factor
lambda05  = sqrt(lambda);
lambdam05 = 1/lambda05;
Wopt      = [1 0.9 0.1 0.2]'; 
MSE       = zeros(K,1);
delta     = 1e-4;
for ens=1:ENS
  disp(ens);
  % Generating input and reference (desired) signals:
  x     = randn(K,1);
  x     = filter(num,den,x); x=x/std(x);  x=x*sqrt(varx);
  noise = randn(K,1); noise=noise/std(noise); noise=noise*sqrt(varn);  
  d     = filter(Wopt,1,x)+noise;
  %----------------
  % Initialization:
  %----------------   
  dfq2=zeros(N+1,1);
  dq2=zeros(N+1,1);
  moduloef=delta;
  a=zeros(N+1,1);
  ctheta=ones(N+1,1);
  cthetaprimeb=ones(N+1,1);
  stheta=zeros(N+1,1);
  sthetaprimeb=zeros(N+1,1);
  invgamma=1;
  %-------------------
  % Runing the filter:
  %-------------------
  for k=1:K
    if rem(k,10000)==0
      disp(k);
    end
    %------------------------------------
    % 1. Obtaining efq1(k+1) and ef(k+1):
    %------------------------------------
    efq1=x(k);
    for i=1:N+1
      oldfriend=dfq2(i);
      dfq2(i)=stheta(i)*efq1+ctheta(i)*lambda05*dfq2(i);
      efq1=ctheta(i)*efq1-stheta(i)*lambda05*oldfriend;
    end % for i
    ef=efq1*invgamma;
    %---------------------
    % 2. Obtaining a(k+1):
    %---------------------
    aNplus2=[a
             ef/(lambda05*moduloef)];
    aux=aNplus2(1);
    for i=1:N+1
      a(i)=sthetaprimeb(i)*aux+cthetaprimeb(i)*aNplus2(i+1);
      aux=cthetaprimeb(i)*aux-sthetaprimeb(i)*aNplus2(i+1);
    end % for i
    %--------------------------
    % 3. Obtaining Qthetaf(k+1)
    %--------------------------
    moduloefold=moduloef;
    % updating ||ef(k)||
    moduloef=sqrt(lambda*moduloef^2+(efq1)^2);
    cthetaf=lambda05*moduloefold/moduloef;
    sthetaf=efq1/moduloef;                  
    %---------------------
    % 4. Obtaining c(k+1):
    %---------------------    
    c=[1
       zeros(N+1,1)];
    % c=Qthetab'*c;
    for i=(N+1):-1:1
      c(i+1)=-sthetaprimeb(i)*c(1);
      c(1)=cthetaprimeb(i)*c(1);
    end
    caux=[0
          c];
    % caux(1:N+2)=Qtheta(k)*caux(1:N+2)
    for i=1:(N+1)
      oldfriend=caux(i+1);
      caux(i+1)=stheta(i)*caux(1)+ctheta(i)*caux(i+1);
      caux(1)=ctheta(i)*caux(1)-stheta(i)*oldfriend;
    end;
    % caux=Qthetaf(k)*caux
    oldfriend=caux(N+3);
    caux(N+3)=sthetaf*caux(1)+cthetaf*caux(N+3);
    caux(1)=cthetaf*caux(1)-sthetaf*oldfriend;
    c=caux(2:(N+3));
    %--------------------------------
    % 5. Obtaining Qthetaprimeb(k+1):
    %--------------------------------
    b=c(1);
    for i=1:(N+1)
      oldfriend=b;
      b=sqrt(b^2+(c(i+1))^2);
      cthetaprimeb(i)=oldfriend/b; 
      sthetaprimeb(i)=-c(i+1)/b;   
    end
    %------------------------
    % 6. Obtendo Qtheta(k+1):
    %------------------------
    invgamma=1;
    for i=1:N+1
      oldfriend=invgamma;
      invgamma=sqrt(invgamma^2+abs(a(i))^2);
      ctheta(i)=oldfriend/invgamma;
      stheta(i)=a(i)/invgamma;
    end; % for i
    %-----------------------------
    % 7. Joint Process Estimation:
    %-----------------------------
    eq1=d(k);
    for i=1:N+1
        oldfriend=eq1;
        eq1=ctheta(i-1+1)*eq1-stheta(i-1+1)*lambda05*dq2(N+2-i);
        dq2(N+2-i)=stheta(i-1+1)*oldfriend+ctheta(i-1+1)*lambda05*dq2(N+2-i);
    end; % for i    
    %------------------------------
    % 8. Updating the output error:
    %------------------------------
    ekaposteriori=eq1/invgamma;
    ekapriori=eq1*invgamma;
    MSE(k)=MSE(k)+abs(ekapriori)^2;
    % MSE(k)=MSE(k)+abs(ekaposteriori)^2;    
  end % for k
end % for ens
MSE=MSE/ENS;
%-------------------
% Simulation result:
%-------------------
E=10*log10(MSE);
k=1:K;
plot(k,E,'k');
xlabel('k'); ylabel('dB');
title('FQR\_PRI\_F');
%--------------------> END OF FQRPRIF.M <--------------------%

