function [E, stheta, ctheta, dq2, dfq2, Ef, f, gamma] = ...
block_mcfqr_pos_b (stheta, ctheta, dq2, dfq2, Ef, f, gamma, X, d, lambda, M, L, pi)
%  +---------------------------------------+
%  |     Block Multichannel FQRD-RLS       |
%  |  Algorithm  (a posteriori version)    |
%  |      First version:    2005 by        |
%  |                                       |
%  |    Apolinário Jr., apolin@ime.eb.br   |
%  |               &                       |
%  |     Werner, stefan.werner@hut.fi      |
%  |               &                       |
%  |    Ramos, alopesramos@yahoo.com.br    |
%  |                                       |
%  |     Last update: Summer 2008 by       |
%  |         Antonio L. L. Ramos           |
%  |       alopesramos@yahoo.com.br        |
%  +---------------------------------------+
% The algorithms have been implemented in a simple way
% (if I can say so...) to make them easy to understand.
% That means that, after you have mastered them, there is a
% lot of simplifications that can be achieved.
% So, feel free to modify and/or adjust them to your own experiments.
%
% It should be clear that the files are meant to be used for educational
% purposes only. No warranty for comercial use is offered and neither the publisher nor
% the authors shall be liable for any loss of any type.



sinBar_f=zeros(1:M);
cosBar_f=zeros(1:M);
    %-------------------------------------+
    % 1 Obtaining Dfq2(k+1) and efq1(k+1):|
    %-------------------------------------+    
    efq1(1,:) = X(1,:);
    for j = 1:L
        for s = 1:M
            temp1 = efq1(s);
            temp2 = lambda^.5 * dfq2 (L - j + 1, s);
            efq1(s) = temp1 * ctheta(j) - temp2 * stheta (j);
            dfq2(L - j + 1, s) = temp1 * stheta(j) + temp2 * ctheta (j);
        end
    end 
    %-----------------------+
    % 2  Obtaining Ef(K+1): |
    %-----------------------+
    temp_efq1 = efq1;
    for i = 1:M
        temp1 = temp_efq1(i);
        temp2 = lambda^.5 * Ef(M - i + 1, i);
        temp3 = sqrt (temp1^2 + temp2^2);
        sinBar_f (i) = temp1/temp3;
        cosBar_f (i) = temp2/temp3;
        Ef (M - i + 1, i) = temp3;
        for s = i + 1:M
            temp1 = temp_efq1(s);
            temp2 = lambda^.5 * Ef (M - i + 1, s);
            temp_efq1(s) = temp1 * cosBar_f(i) - temp2 * sinBar_f(i);
            Ef (M - i + 1, s) = temp1 * sinBar_f(i) + temp2 * cosBar_f(i);
        end
    end
    %--------------------+
    % 3 Obtaining p(k+1):|
    %--------------------+    
    temp_gamma = gamma;
    for b = 1:M
        p(M - b + 1, 1) = sinBar_f(b)*temp_gamma;
        temp_gamma      = cosBar_f(b)*temp_gamma;
    end
    %--------------------------+
    % 4 Obtaining Q'theta_f(k+1):|
    %--------------------------+    
    Ef_0 = Ef;
    temp_dfq2 = dfq2;
    for j = 1:L
        for s = 1:M
            if j<=(L-p(s)+s)
                temp1 = temp_dfq2 (j, s);
                temp2 = Ef_0 (M - s + 1, s);
                sintheta_prime_f (j, s) = temp1 / sqrt (temp1^2 + temp2^2);
                costheta_prime_f (j, s) = temp2 / sqrt (temp1^2 + temp2^2);
                Ef_0 (M - s + 1, s) = sqrt (temp1^2 + temp2^2);
                for b = s + 1:M
                    temp1 = temp_dfq2 (j, b);
                    temp2 = Ef_0 (M - s + 1, b);
                    temp_dfq2 (j, b) = temp1 * costheta_prime_f(j, s) - temp2 * sintheta_prime_f(j, s);
                    Ef_0 (M - s + 1, b) = temp1 * sintheta_prime_f(j, s) + temp2 * costheta_prime_f(j, s);
                end
            end
        end
    end
    %---------------------+
    % 5 Obtaining fN(k+1):|
    %---------------------+    
	for j = 1:L
        for s = 1:M
            if j<=(L-pi(s)+s)
                temp1 = f (M + j);
                temp2 = p(M - s + 1);
                f (j) = temp1 * costheta_prime_f (j, s) - temp2 * sintheta_prime_f (j, s);
                f (M + j) = temp1 * costheta_prime_f (j, s) - temp2 * sintheta_prime_f (j, s);
                p (M - s + 1) = temp1 * sintheta_prime_f (j, s) + temp2 * costheta_prime_f (j, s);
            end
        end
	end
    f(L+1:L + M) = p;      
    %-----------------------------------------------------------------------------
    % Performing the desired permutations    
    for s = M:-1:1
         aux=f(L+M-s+1);
         f(L+M-pi(s)+2:L+M-s+1)=f(L+M-pi(s)+1:L+M-s);
         f(L+M-pi(s)+1)=aux;
    end
    %----------------------------------------+
    % 6 Obtaining Qtheta(k+1) and gamma(k+1):|
    %----------------------------------------+    
	gamma = 1;
	for j = 1:L
            stheta(j) = f(L+M-j+1)/gamma;
            ctheta(j) = sqrt (1 - stheta(j)^2);
            gamma = gamma * ctheta (j);
	end
    %-------------------+
    % 7 Join Estimation:|
    %-------------------+    
	eq1 = d;
	for j = 1:L
        temp1 = eq1;
        temp2 = lambda^.5 * dq2(L - j + 1);
        eq1 = temp1 * ctheta (j) - temp2 * stheta(j);
        dq2(L - j + 1) = temp1 * stheta (j) + temp2 * ctheta(j);
	end
	E = eq1 / (realmin + gamma);    % a priori error
	% E = eq1 * gamma;          % a posteriori error   
