%  +--------------------------------------+
%  |    First version  Spring 2004 by:    |
%  |         Antonio L. L. Ramos          |
%  |     Last update: Summer 2008 by:     |
%  |         Antonio L. L. Ramos          |
%  |       alopesramos@yahoo.com.br       |
%  +--------------------------------------+
%=====================================================+
% This is the main file implementing the application  |
% example discribed Chapter 6, Section 6.6, of the    |
% book 'QRD-RLS Adaptive Filtering (2009)', edited by |
% José Antonio Apolinário Jr.                         |
%=====================================================+
% 
% The algorithms have been implemented in a simple way
% (if I can say so...) to make them easy to understand.
% That means that, after you have mastered them, there is a
% lot of simplifications that can be achieved.
% So, feel free to modify and/or adjust them to your own experiments.
%
% It should be clear that the files are meant to be used for educational
% purposes only. No warranty for comercial use is offered and neither the publisher nor
% the authors shall be liable for any loss of any type.

clear, clc,
close all
ENS = 50% 100;%200 % Number of Ensembles (Independent runs)
kmax = 1000 % Max Number of Samples
%----------------------------------------------------------------------------------------------
% Problem stars here...
M = 5; % Number of channels
channels = [4; 4; 3; 2; 1]; % Length of each channel
N       = channels(1);
P = sum(channels); % Sum of channels length
X = zeros(1, M);   % A vector containing the most recent sample of each channel

gamma_block_pri = 1;
gamma_block_pos = 1;
gamma_seq_pri   = 1;
gamma_seq_pos   = 1;
lambda          = .98; % Forgetting factor
%----------------------------------------------------------------------------------------------

%--------------------------------------
% Implementing Eq 6.9, pp 140:
%--------------------------------------
p = zeros(M+1,1);
for i=1:M
    p(i)= 0;
    for r=1:i-1 
        p(i)=p(i)+r*(channels(r)-channels(r+1)); % vector channels contains
    end                         % each channel length
    p(i)=p(i)+i;
end    
% Or equivalently for this particular case:
% m =[1; 2; 5; 9; 14; 0];

%-------------------------------------------------------
% Signal and noise statistics:
varx = 1;
varn = 1e-6;
%-------------------------------------------------------
% The optimal weight vector:
wopt = [-0.78; -1.48; -1.39; 0.04; 0.54; 3.72; 1.86; -0.76; -1.62; 0.76; -0.12; 1.41; -1.52; -0.13];
%-------------------------------------------------------
% Initializing the MSE vectors:
%----------------------------------------
% Sequential Multichannel algorithms... 
MSE_seq_pri  = zeros(kmax, 1);
MSE_seq_pos  = zeros(kmax, 1);
% Block Multichannel algorithms...
MSE_block_pri  = zeros(kmax, 1);
MSE_block_pos  = zeros(kmax, 1);

%---------------------------------------------------
% If you wish to try other algorithms...
% (Some of these are discribed or refered to in the book)
W_InvQRD_ens=zeros(P,1);  % Inverse QRD-RLS Algorithm
W_RLS_ens   = zeros(P,1); % RLS Algorithm
W_NLMS_ens  = zeros(P,1); % Normalized LMS Algorithm
W_LMS_ens   = zeros(P,1); % LMS Algorithm
W_CG_ens   = zeros(P,1);  % Conjugate Gradient Algorithm 
%-------------------------------------------------------
% Initializing their respective MSE vectors:
%--------------------------------------------
MSE_RLS  = zeros(kmax, 1);
MSE_InvQRD  = zeros(kmax, 1);
MSE_NLMS  = zeros(kmax, 1);
MSE_LMS  = zeros(kmax, 1);
MSE_CG  = zeros(kmax, 1);
%----------------------------------------------------------------------------------------------
for j=1:ENS
	    
    disp(j);
    %-------------------------------------------+
    %  Initializations for each independent run:|
    %-------------------------------+-----------+
    % Block Multichannel Algorithms:|
    %-------------------+-----------+    
    % block_mcfqr_pos_b:|
    %-------------------+    
    efq1_block_pos = zeros (1, M);
    dfq2_block_pos = zeros (P, M);
    f_block = zeros (P+ M, 1);
    Ef_block_pos = zeros (M);
    for i =1:M
        Ef_block_pos(M-i+1, i) = 1;
    end
    Ef_0_block_pos = zeros (M);
    for i =1:M
        Ef_0_block_pos(M-i+1, i) = 1;
    end
    eq1_block_pos = 1;
    dq2_block_pos = ones (P, 1);
    % sintheta_prime_block_pos = zeros (P, M);
    % costheta_prime_block_pos = zeros (P, M);
    costheta_block_pos = ones (P, 1);
    sintheta_block_pos = zeros (P, 1);
    %-------------------+    
    % block_mcfqr_pri_b:|
    %-------------------+    
    efq1_block_pri = zeros (1, M);
    dfq2_block_pri = zeros (P, M);
    a_block = zeros (P+ M, 1);
    Ef_block_pri = zeros (M);
    for i =1:M
        Ef_block_pri(M-i+1, i) = 1;
    end
    Ef_0_block_pri = zeros (M);
    for i =1:M
        Ef_0_block_pri(M-i+1, i) = 1;
    end
%     inv_gamma_block_pri = 1;
    eq1_block_pri = 1;
    dq2_block_pri = ones (P, 1);
    sintheta_prime_block_pri = zeros (P, M);
    costheta_prime_block_pri = ones (P, M);
    sintheta_block_pri = zeros (P, 1);
    costheta_block_pri = ones (P, 1);
    %------------------------------------+
    % Sequential Multichannel Algorithms:|
    %-----------------+------------------+    
    % seq_mcfqr_pri_b:|
    %-----------------+    
    efq1_seq_pri        = zeros (1, M);
	dfq2_seq_pri        = zeros (P, M);
	a_seq               = zeros (P+ 1, 1);
	ef_seq_pri          = ones(M, 1);
	eq1_seq_pri         = 1;
	dq2_seq_pri         = zeros (P, 1);
	sen_p_seq_pri       = zeros (P, M);
	cos_p_seq_pri       = ones (P, M);
	coseno_seq_pri      = ones (P, 1);
	seno_seq_pri        = zeros (P, 1);
    %-----------------+
    % seq_mcfqr_pos_b:|
    %-----------------+    
    efq1_seq_pos        = zeros (1, M);
	dfq2_seq_pos        = zeros (P, M);
	f_seq               = zeros (P + 1, 1);
	ef_seq_pos          = ones(M, 1);
	eq1_seq_pos         = 1;
	dq2_seq_pos         = zeros (P, 1);
	sen_p_seq_pos       = zeros (P, M);
	cos_p_seq_pos       = ones (P, M);
	coseno_seq_pos      = ones (P, 1);
	seno_seq_pos        = zeros (P, 1);
    %------------------+
    % Other Algorithms:|
    %------------------+---------------------------------+
    % Obs:                                               |
    % You might have to check for any minor bug on these |
    % algorithms. I did not perform a last check on them.|
    % Good luck!                                         |
    % Antonio                                            |
    %----------------------------------------------------+
    % RLS:|
    %-----+
    W_RLS=zeros(P,1);
    invR_RLS=eye(P);
    %------------+
    % InvQRD-RLS:|
    %------------+    
    W_InvQRD = zeros(P,1);
    InvU =eye(P);
    %------+
    % NLMS:|
    %------+    
    W_NLMS=zeros(P,1);
    %-----+
    % LMS:|
    %-----+    
    W_LMS=zeros(P,1);
    %--------------------+
    % Conjugate Gradient |
    %--------------------+
    W_CG   = zeros(P,1);  
    R_CG   = eye(P);
    gCG       = zeros(P, 1);
    cCG       = gCG;
    eta     = lambda;
    %-----------------------------+
    % Generating the imput signal:|
    %-----------------------------+    
    xx = zeros(N, 1);
    xxx = zeros(P,1);
    x = randn(kmax, 1);% + i*randn(kmax, 1); 
    % You can change the line above to generating a complex signal
    % Note however that you might have to do some modifications on the Block Mutichannel functions
    % because, unlike their Sequential counterparts, the Multichannel Block
    % algorithms are not implemted in the complex form.
    x = x/mean(x);  
    x = x/std(x)*sqrt(varx);
    %--------------------------------------------------------------------
    % Transforming the white noise into a colored noise...
    x = filter(1,[1 -0.9],x); %
    %--------------------------------------------------------------------
    n = randn(kmax, 1);% + i*randn(kmax, 1);      % Gerando o ruido...
    n = n/mean(n);
    n = n/std(n)*sqrt(varn);
    
    for k=1:kmax
        xx   = [x(k); xx(1: N-1)];
        % Obtaining the most recent sample of each channel...
		X(1, 1)    = x(k);
		X(1, 2)    = x(k)*x(k);
		X(1, 3)    = x(k)*xx(2);
		X(1, 4)    = x(k)*xx(3);
		X(1, 5)    = x(k)*xx(4);
        %--------------------------------------+
        % Generating the Volterra imput vector:|  
        %--------------------------------------+
        xxx(1)   = x(k);
        xxx(2)   = x(k)^2;

        xxx(3)   = xx(2);
        xxx(4)   = xx(2)^2;
        xxx(5)   = x(k)*xx(2);

        xxx(6)   = xx(3);
        xxx(7)   = xx(3)^2;
        xxx(8)   = xx(2)*xx(3);
        xxx(9)   = x(k)*xx(3);

        xxx(10)  = xx(4);
        xxx(11)  = xx(4)^2;
        xxx(12)  = xx(3)*xx(4);
        xxx(13)  = xx(2)*xx(4);
        xxx(14)  = x(k)*xx(4);
        %------------------------------+ 
        % Obtaining the desired output:|
        %------------------------------+
        d = wopt'*xxx + n(k);
        %----------------------------------------------------
        % Complex versions:
        %----------------------------------------------------
        % Multichannel Sequential Algorithms:
        [E_seq_pri, seno_seq_pri, coseno_seq_pri, sen_p_seq_pri, cos_p_seq_pri, dq2_seq_pri, dfq2_seq_pri, ef_seq_pri, a_seq, gamma_seq_pri] = ...
        seq_mcfqr_pri_b (seno_seq_pri, coseno_seq_pri, sen_p_seq_pri, cos_p_seq_pri, dq2_seq_pri, dfq2_seq_pri, ef_seq_pri, a_seq, gamma_seq_pri, X, d, lambda, M, P, p);
  
        [E_seq_pos, seno_seq_pos, coseno_seq_pos, sen_p_seq_pos, cos_p_seq_pos, dq2_seq_pos, dfq2_seq_pos, ef_seq_pos, f_seq, gamma_seq_pos] = ...
        seq_mcfqr_pos_b (seno_seq_pos, coseno_seq_pos, sen_p_seq_pos, cos_p_seq_pos, dq2_seq_pos, dfq2_seq_pos, ef_seq_pos, f_seq, gamma_seq_pos, X, d, lambda, M, P, p);
        %------------------------------------------------------------------
  
        % Multichannel Block Algorithms:
        [E_block_pri, sintheta_block_pri, costheta_block_pri, sintheta_prime_block_pri, costheta_prime_block_pri, dq2_block_pri, dfq2_block_pri, Ef_block_pri, a_block, gamma_block_pri] = ...
        block_mcfqr_pri_b (sintheta_block_pri, costheta_block_pri, sintheta_prime_block_pri, costheta_prime_block_pri, dq2_block_pri, dfq2_block_pri, Ef_block_pri, a_block, gamma_block_pri, X, d, lambda, M, P, p);

        [E_block_pos, sintheta_block_pos, costheta_block_pos, dq2_block_pos, dfq2_block_pos, Ef_block_pos, f_block, gamma_block_pos] = ...
        block_mcfqr_pos_b (sintheta_block_pos, costheta_block_pos, dq2_block_pos, dfq2_block_pos, Ef_block_pos, f_block, gamma_block_pos, X, d, lambda, M, P, p);
    
%-------------------+
% Others Algorithms:|
%-------------------+
% RLS:
% [W_RLS, E_RLS, invR_RLS]=RLS(W_RLS, invR_RLS, d, xxx, lambda);
% InvQRD-RLS:
% [W_InvQRD, E_InvQRD, InvU]=InvQRD_RLS(W_InvQRD, InvU, xxx, d, lambda, L);
% NLMS:
% function [W, E]=NLMS(W,      d, X, mu, gamma);
% [W_NLMS, E_NLMS]=NLMS(W_NLMS,  d, xxx, 0.12);

% [W_NLMS, E_NLMS]=nlms(W_NLMS,  d, xxx, 0.12);
[W_NLMS, E_NLMS]=nlms(W_NLMS,  d, xxx, 1);


% LMS:
% function [W, E]=LMS(W,      d, X, mu);
% [W_LMS, E_LMS]=LMS(W_LMS,  d, xxx, 0.008);
 

% W_RLS_ens = W_RLS_ens+W_RLS;
% W_InvQRD_ens = W_InvQRD_ens+W_InvQRD;
W_NLMS_ens = W_NLMS_ens+W_NLMS;
% W_LMS_ens = W_LMS_ens+W_LMS;
% function [W, E, R_CG, c, g]=CG(W, c, g, R_CG, d, X, lambda, eta);
% [W_CG, E_CG, R_CG, cCG, gCG]=CG(W_CG, cCG, gCG, R_CG, d, xxx, lambda, eta);


MSE_seq_pos(k)  = MSE_seq_pos(k) + abs(E_seq_pos)^2;
MSE_seq_pri(k)  = MSE_seq_pri(k) + abs(E_seq_pri)^2;
MSE_block_pos(k)  = MSE_block_pos(k) + abs(E_block_pos)^2;
MSE_block_pri(k)  = MSE_block_pri(k) + abs(E_block_pri)^2;

% MSE_RLS(k)  = MSE_RLS(k) + abs(E_RLS)^2;  
% MSE_InvQRD(k)  = MSE_InvQRD(k) + abs(E_InvQRD)^2;  
MSE_NLMS(k)  = MSE_NLMS(k) + abs(E_NLMS)^2;  
% MSE_LMS(k)  = MSE_LMS(k) + abs(E_LMS)^2;  
% MSE_CG(k)  = MSE_CG(k) + abs(E_CG)^2;  
    end
end
% W_RLS = W_RLS_ens/ENS;
% W_InvQRD = W_InvQRD_ens/ENS;
W_NLMS = W_NLMS_ens/ENS;
% W_LMS = W_LMS_ens/ENS;

MSE_seq_pri  = MSE_seq_pri/ENS;
MSE_seq_pos  = MSE_seq_pos/ENS;

MSE_block_pri  = MSE_block_pri/ENS;
MSE_block_pos  = MSE_block_pos/ENS;

% MSE_RLS  = MSE_RLS/ENS;
 
%  MSE_InvQRD  = MSE_InvQRD/ENS;
MSE_NLMS  = MSE_NLMS/ENS;
% MSE_LMS  = MSE_LMS/ENS;
%  MSE_CG  = MSE_CG/ENS;
%---------------------+
% Ploting the results:|
%---------------------+
figure, hold
%-------------------------+
% Sequential Multichannel:|
%-------------------------+
plot(10*log10(MSE_seq_pos),'k')
plot(10*log10(MSE_seq_pri), '--r')
%--------------------+
% Block Multichannel:|
%--------------------+
plot(10*log10(MSE_block_pri), '--g')
plot(10*log10(MSE_block_pos),':b')
%------+
% NLMS:|
%------+
plot(10*log10(MSE_NLMS), ':k')

plot(10*log10(varn*ones(kmax, 1)), '--k')
text(700,10,'NLMS')
text(750,-55,'MCFQR Algorithms')
xlabel('Sample')
ylabel('MSE (dB)')
axis([1 1000 -70 20])

% Ploting of other results if needed...
% plot(10*log10(MSE_RLS), ':r')    % RLS
% plot(10*log10(MSE_InvQRD), '--b')    % InvQRD-RLS
% plot(10*log10(MSE_LMS), '--k')    % LMS
% plot(10*log10(MSE_CG), ':g')    % CG

