function [E, steta, cteta, stetalinhaf, ctetalinhaf, dq2, dfq2, ef, a, gamma] = ...
seq_mcfqr_pri_b (steta, cteta, stetalinhaf, ctetalinhaf, dq2, dfq2, ef, a, gamma, X, d, lambda, M, L, pi)
%  +--------------------------------------+
%  |   Sequential Multichannel FQRD-RLS   |
%  |    Algorithm  (a priori version)     |
%  |    First version  Spring 2004 by:    |
%  |         Antonio L. L. Ramos          |
%  |                                      |
%  |     Last update: Summer 2008 by:     |
%  |         Antonio L. L. Ramos          |
%  |       alopesramos@yahoo.com.br       |
%  +--------------------------------------+
% The algorithms have been implemented in a simple way
% (if I can say so...) to make them easy to understand.
% That means that, after you have mastered them, there is a
% lot of simplifications that can be achieved.
% So, feel free to modify and/or adjust them to your own experiments.
%
% It should be clear that the files are meant to be used for educational
% purposes only. No warranty for comercial use is offered and neither the publisher nor
% the authors shall be liable for any loss of any type.


efq1(1, :) = X(1,:);
for i = 1:M
	% Step 1: Updating dfq2(k-1)
    %------------------------------------+
    % 1. Otaining dfq2(j ,i) e efq1(1, i):|
    %------------------------------------+
	for j = 1:L % 1:k
      oldfriend     = efq1(1, i);
      efq1(1, i)    = cteta(j)*efq1(1, i) + steta(j)*lambda^0.5*dfq2(L-j+1, i);
      dfq2(L-j+1, i)= cteta(j)*lambda^0.5*dfq2(L-j+1, i) - conj(steta(j))*oldfriend;
    end;
    r= conj(efq1(i))/(gamma*lambda^0.5*ef(i)) ; % Pode ser desta forma ou como em # 3. 
                                                    % apos obter os angulos em # 2.
    %-------------------------+
	% 2. Otaining ||ef(k+1)||: |
	%-------------------------+
% 	temp    = lambda^.5*ef(i);
    temp    = ef(i);
    ef(i)   = sqrt( abs(efq1(i))^2+ lambda*abs(ef(i))^2);
    cos_pp  = abs(temp)/ef(i);
    sen_pp  = conj(cos_pp * efq1(i)/temp);
    %--------------------+
	% 3. Otaining a(k+1): |
	%--------------------+
    % aux = (sen_pp/gamma)/cos_pp;
    for j = L:-1:pi(i),
        temp2       = a(L-j+1+1);
        a(L-j+1)    = ctetalinhaf(j, i)*temp2 - conj(stetalinhaf(j, i))*r;
        r           = stetalinhaf(j, i)*temp2 + ctetalinhaf(j, i)*r;
    end  
    a(L+1 - pi(i)+1) = r;
	%--------------------------+
	% 4. Otaining Qteta'f(k+1): |
	%--------------------------+
    % step 3: Obtaining Qthetaf'(k+1)
	efi_0 = ef(i);    
    for j = L:-1:pi(i),
      temp = efi_0;
      efi_0 = sqrt(efi_0^2 + abs(dfq2( L-j+1, i))^2);
      ctetalinhaf(j, i) = abs(temp/ efi_0);
      stetalinhaf(j, i) = conj(ctetalinhaf(j, i)*dfq2(L-j+1, i)/temp);
    end; 
    %-----------------------+
	% 5. Otaining Qteta(K+1) |
	%-----------------------+
    if (i == 1)
        invgamma = 1;
    else
        invgamma = invgamma_old;
    end
    for j = pi(i):L 
        temp = invgamma;
        invgamma  = sqrt(invgamma^2 + abs( a( L-j +1 +1 ) )^2);        
        
         if (j ==  pi(i+1)-1)
             invgamma_old = invgamma;
         end
        cteta(j)= abs(temp)/invgamma;
        steta(j)= conj( -a( L-j+1 +1) * cteta( j)/temp);
        gamma       = 1/invgamma;
    end
end % i loop
%------------------------------+
% 6. Joint Process Estimation: |
%------------------------------+
eq1 = d;
for j=1:L
  oldfriend     = eq1;
  eq1           = cteta(j)*eq1+steta(j)*lambda^0.5*dq2(L-j+1);
  dq2(L-j+1)    = cteta(j)*lambda^0.5*dq2(L-j+1) - conj(steta(j))*oldfriend;
end
%-------------------------------+
% 7. Updating the output error: |
%-------------------------------+
E = eq1/(realmin + gamma);% a priori error
% E = eq1*gamma;            % a posteriori error
