function e = be_iqrd_rls(y,x,h,l,M,Q);
%
% function e = be_iqrd_rls(y,x,h,l,M,Q);
%
% The block exact Householder inverse QRD-RLS algorithm
%
% y: channel output vector
% x: channel input vector
% h: channel impulse response vector
% l: forgetting factor
% M: size of data 
% Q: the block size
%
% e: squared weight error vector

%
%Initialization
%
N= length(h);
delta=1*l^N;
U=inv(sqrt(delta))*eye(N);
dq = zeros(N,1);
e=[];
X = zeros(M,N);
d = zeros(M,1);
Y = zeros(N+Q,N);
ww = zeros(N,1);
%
% Construction of the data matrices
%
for m=N:M+N-1,
 u=[x(m:-1:m-N+1)];
 X(M+N-m,:) = u;
 d(M+N-m) =  y(m-N+1);
end;
%
% The algorithm
%
L = diag(l.^([0:Q-1]/2));
w = zeros(N,M/Q);
for i=1:M/Q,
    Xt = L*X(M-i*Q+1:M-(i-1)*Q,:);
    dt = L*d(M-i*Q+1:M-(i-1)*Q);
    G = -l^(-Q/2)*U*Xt.';
    Ga = [eye(Q);G];
    Y = [zeros(Q,N); l^(-Q/2)*U];
    for k=1:N,
        [Z,H] = roht(Ga,Q+k,1);
        Y = H*Y;
        Ga = Z;
    end;
    U = Y(Q+1:Q+N,:);
    D = Ga(1:Q,:);
    E = Y(1:Q,:);
    ww = ww -E'*inv(D')*(conj(dt) - conj(Xt)*ww);  
    w(:,i)= ww;
end;

for i=1:M/Q,
    e(i) = norm(w(:,i)-h')^2;
end;
figure
semilogy(e)
    