function e = be_qrd_rls(y,x,h,l,M,Q);
%
% function e = be_qrd_rls(y,x,h,l,M,Q);
%
% The block exact Householder QRD-RLS algorithm
%
% y: channel output vector
% x: channel input vector
% h: channel impulse response vector
% l: forgetting factor
% M: size of data 
% Q: the block size
%
% e: squared weight error vector

%
%Initialization
%
N = length(h);
delta=1*l^N;
U=sqrt(delta)*eye(N);
dq = zeros(N,1);
e=[];
X = zeros(M,N);
d = zeros(M,1);
Y = zeros(N+Q,N);
%
% Construction of the data matrices 
%
for m=N:M+N-1
 u=[x(m:-1:m-N+1)];
 X(M+N-m,:) = u;
 d(M+N-m) =  y(m-N+1);
end;
%
% The algorithm
%
L = diag(l.^([0:Q-1]/2));
w = zeros(N,M/Q);
for i=1:M/Q,
    Y = [L*conj(X(M-i*Q+1:M-(i-1)*Q,:)) ; l^(Q/2)*U];
    g = [L*conj(d(M-i*Q+1:M-(i-1)*Q));l^(Q/2)*dq];
    for k=1:N,
        v = Y(:,k);
        [z,H] = oht(v,Q+k,1,Q);
        Y(:,k:N) = H*Y(:,k:N);
        g = H*g;
    end;
    U = Y(Q+1:Q+N,:); 
    dq = g(Q+1:Q+N);
    w(:,i)= inv(U)*dq;
end;

for i=1:M/Q,
    e(i) = norm(w(:,i)-h')^2;
end;
figure
semilogy(e)
    
        
    
    