function [z,H] = hht(x,i,phi);
%
% function [z,H] = hht(x,i,phi)
%
% Construction and application of the hyperbolic Householder transform to 
% vector x. All hyperbolic energy of x is compressed to its i-th enrty. 
% The i-th element of phi must be +1. 
%
% phi: vector containing the diagonal elements of the hyperbolic transform 
%      signature 
% H: the resulting hyperbolic Householder transform
% z: the resulting sparse vector

N = length(x);
y = zeros(N,1);
y = diag(phi)*x;
y(i) = y(i) + (x(i)/abs(x(i)))*sqrt(x'*diag(phi)*x);
b = 2/(y'*diag(phi)*y);
H = diag(phi)-b*y*y';
z = H*x;

