function [z,H] = oht(x,i,p,q);
%
% function [z,H] = oht(x,i,p,q)
%
% Construction and application of the orthogonal Householder transform 
% to vector x.
% The elements x(p:q) are zeroed with respect to the i-th element of x. 
% For complex entries the i-th element of x must be real. 
%
% H: the resulting Householder matrix
% z: the resulting sparse vector

N = length(x);
I = eye(N);
y = zeros(N,1);
y(p:q) = x(p:q);
y(i) = x(i) + norm([x(p:q);x(i)]);
b = 2/(norm(y)^2);
H = I-b*y*y';
z = H*x;

