function [Z,H] = rhht(X,phi,i,p);
%
% function [Z,H] = rhht(X,phi,i,p)
%
% Generation of a row hyperbolic Householder transform and its 
% application to a tall matrix X. The i-h row of X is zeroed with 
% respect to a square matrix starting at row p of X. 
% The i-th element of phi must be +1.
%
% phi: vector containing the diagonal elements of the hyperbolic transform 
%      signature
% H: The resulting row hyperbolic Householder transform
% Z: The resulting matrix, Z = H*X
%
m = length(X(:,1)); 
y = zeros(m,1);
I = eye(m);
n = length(X(1,:));
phi1 = phi(p:p+n-1); 
u = inv(X(p:p+n-1,:))'*X(i,:)';
y(p:p+n-1) = u;
y(i) = -1-sqrt(1+phi(i)*u'*diag(phi1)*u);
y = y/(norm(X(i,:)));
b = 2/(y'*diag(phi)*y);
H = diag(phi)-b*y*y';
Z = H*X; 