function [Z,H] = roht(X,i,p);
%
% function [Z,H] = roht(X,i,p)
%
% Generation of a row orthogonal Householder transform and its 
% application to a tall matrix X. The i-h row of X is zeroed with 
% respect to a square matrix starting at row p of X. 
%
% H: The resulting row orthogonal Householder transform
% Z: The resulting matrix, Z = H*X
%
y = zeros(length(X(:,1)),1);
I = eye(length(X(:,1)));
n = length(X(1,:));
u = inv(X(p:p+n-1,:))'*X(i,:)';
y(p:p+n-1) = u;
y(i) = -1-sqrt(1+u'*u);
y = y/(norm(X(i,:)));
b = 2/(norm(y)^2);
H = I-b*y*y';
Z = H*X; 
